import webbrowser


def open_contact() -> None:
    """
    Opens the contact page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_documentation() -> None:
    """
    Opens the documentation on read-the-docs for PROBE
    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_donate() -> None:
    """
    Opens the donation page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_examples() -> None:
    """
    Opens the examples page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_github() -> None:
    """
    Opens the GitHub repository for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_issues() -> None:
    """
    Opens the issues page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_pypi() -> None:
    """
    Opens the PyPi page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")
